@@ -2,7 +2,7 @@
 from reahl.sqlalchemysupport import SqlAlchemyControl
 
 reahlsystem.root_egg = 'reahl-webdev'
-reahlsystem.connection_uri = 'postgresql://rhug:rhug@localhost/rhug'
+reahlsystem.connection_uri = 'postgresql:///reahl'
 #reahlsystem.connection_uri = 'sqlite:////tmp/test.db'
 #reahlsystem.connection_uri = 'sqlite:///:memory:'
 reahlsystem.orm_control = SqlAlchemyControl(echo=False)
