@@ -1,4 +1,4 @@
-# Copyright 2013-2016 Reahl Software Services (Pty) Ltd. All rights reserved.
+# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
 #-*- encoding: utf-8 -*-
 #
 #    This file is part of Reahl.
@@ -19,7 +19,7 @@ from __future__ import print_function, unicode_literals, absolute_import, divisi
 
 import six
 import io
-import time
+import re
 import contextlib
 from six.moves.urllib import parse as urllib_parse
 import logging
@@ -33,7 +33,6 @@ from selenium.webdriver.common.keys import Keys
 from selenium.webdriver.common.action_chains import ActionChains
 
 from reahl.component.py3compat import ascii_as_bytes_or_str
-from reahl.component.decorators import deprecated
 from reahl.web.fw import Url
 
 
@@ -57,10 +56,6 @@ def patch_Field():
 
 
 class BasicBrowser(object):
-    def save_source(self, filename):
-        with io.open(filename, 'w') as output:
-            for line in html.tostring(self.lxml_html, pretty_print=True, encoding='unicode').split('\n'): 
-                output.write(line+'\n')
 
     def view_source(self):
         for line in html.tostring(self.lxml_html, pretty_print=True, encoding='unicode').split('\n'): 
@@ -69,11 +64,7 @@ class BasicBrowser(object):
     def save_source(self, filename):
         with io.open(filename, 'w') as html_file:
             html_file.write(self.raw_html)
-            
-    def get_html_for(self, locator):
-        xpath = six.text_type(locator)
-        return html.tostring(self.lxml_html.xpath(xpath)[0], encoding='unicode')
-        
+
     def is_element_present(self, locator):
         xpath = six.text_type(locator)
         return len(self.lxml_html.xpath(xpath)) == 1 
@@ -320,7 +311,44 @@ class Browser(BasicBrowser):
                 form[select.attrib['name']] = list(option.values())[0]
                 return
         raise AssertionError('Option %s not found' % label_to_choose)
-   
+
+    def select_many(self, locator, labels_to_choose):
+        """Finds the select element indicated by `locator` and selects the options labelled as such.
+
+           :param locator: An instance of :class:`XPath` or a string containing an XPath expression.
+           :param labels_to_choose: The labels of the options that should be selected.
+        """
+        xpath = six.text_type(locator)
+        select = self.xpath(xpath)
+        assert len(select) == 1, 'Could not find one (and only one) element for %s' % locator
+        select = select[0]
+        assert select.tag == 'select', 'Expected %s to find a select tag' % locator
+
+        form = self.get_form_for(xpath)
+
+        options_to_select = []
+        for option in select.findall('option'):
+            if option.text in labels_to_choose:
+                options_to_select.append(option)
+        form[select.attrib['name']] = [option.values()[0] for option in options_to_select]
+
+        if len(options_to_select) != len(options_to_select):
+            raise AssertionError('Could only select options labelled[] not all of []' % (','.join([option.text for option in options_to_select]), ','.join(labels_to_choose) ))
+
+    def select_none(self, locator):
+        """Finds the select element indicated by `locator` and ensure nothing is selected.
+
+           :param locator: An instance of :class:`XPath` or a string containing an XPath expression.
+        """
+        xpath = six.text_type(locator)
+        select = self.xpath(xpath)
+        assert len(select) == 1, 'Could not find one (and only one) element for %s' % locator
+        select = select[0]
+        assert select.tag == 'select', 'Expected %s to find a select tag' % locator
+
+        form = self.get_form_for(xpath)
+        form[select.attrib['name']] = []
+
     def get_value(self, locator):
         """Returns the value of the input indicated by `locator`.
 
@@ -402,7 +430,7 @@ class XPath(object):
     @classmethod
     def label_with_text(cls, text):
         """Returns an XPath to find an HTML <label> containing the text in `text`."""
-        return cls('//label[node()="%s"]' % text)
+        return cls('//label[normalize-space()=normalize-space("%s")]' % text)
 
     @classmethod
     def heading_with_text(cls, level, text):
@@ -414,6 +442,11 @@ class XPath(object):
         """Returns an XPath to find an HTML <caption> matching the text in `text`."""
         return cls('//caption[node()="%s"]' % (text))
 
+    @classmethod
+    def option_with_text(cls, text):
+        """Returns an XPath to find an HTML <option> containing the text in `text`."""
+        return cls('//option[node()="%s"]' % (text))
+
     @classmethod
     def table_with_summary(cls, text):
         """Returns an XPath to find an HTML <table summary='...'> matching the text in `text` in its summary attribute value."""
@@ -432,7 +465,7 @@ class XPath(object):
     @classmethod
     def link_with_text(cls, text, nth=1):
         """Returns an XPath to find an HTML <a> containing the text in `text`."""
-        return cls('(//a[normalize-space(node())=normalize-space("%s")])[%s]' % (text, nth))
+        return cls('(//a[normalize-space(.)=normalize-space("%s")])[%s]' % (text, nth))
 
     @classmethod
     def link_starting_with_text(cls, text):
@@ -444,6 +477,11 @@ class XPath(object):
         """Returns an XPath to find an HTML <p> that contains the text in `text`."""
         return cls('//p[contains(node(), "%s")]' % text)
 
+    @classmethod
+    def input_named(cls, name):
+        """Returns an XPath to find an HTML <input> with the given name."""
+        return '//input[@name="%s"]' % name
+
     @classmethod
     def input_labelled(cls, label):
         """Returns an XPath to find an HTML <input> referred to by a <label> that contains the text in `label`."""
@@ -461,12 +499,6 @@ class XPath(object):
         """Returns an XPath to find an HTML <input> with type attribute `input_type`."""
         return '//input[@type="%s"]' % input_type
 
-    @deprecated('Please use fieldset_with_legend() instead.', '3.2')
-    @classmethod
-    def inputgroup_labelled(cls, label):
-        """Returns an XPath to find an InputGroup with label text `label`."""
-        return cls('//fieldset[label[normalize-space(node())=normalize-space("%s")]]' % label)
-
     @classmethod
     def fieldset_with_legend(cls, legend_text):
         """Returns an XPath to find a FieldSet with the given `legend_text`.
@@ -678,8 +710,41 @@ class DriverBrowser(BasicBrowser):
         """Waits for the current page to load."""
         self.wait_for(self.is_page_loaded)
         try:
-            # Turn all jquery animations off for testing
-            self.web_driver.execute_script('if ( "undefined" !== typeof jQuery) { jQuery.fx.off=true; }; return true;')  
+#              styleEl.textContent = '*{ transition: none !important; transition-property: none !important; transform: none !important; animation: none !important;  }';
+            # Turn all jquery and bootstrap animations off for testing
+#            self.web_driver.execute_script('if ( "undefined" !== typeof jQuery) { jQuery.fx.off = true; jQuery.support.transition = false; }; return true;')
+            self.web_driver.execute_script('''
+              if ( "undefined" !== typeof jQuery) { jQuery.fx.off = true; jQuery.support.transition = false; };
+              var styleEl = document.createElement('style');
+              styleEl.textContent = '*{ ' +
+                 'transition-delay: 0s  !important; ' +
+                 '-o-transition-delay: 0s !important; ' +
+                 '-moz-transition-delay: 0s !important; ' +
+                 '-ms-transition-delay: 0s !important; ' +
+                 '-webkit-transition-delay: 0s !important; ' +
+
+                 'transition-duration: 0s !important; ' +
+                 '-o-transition-duration: 0s !important; ' +
+                 '-moz-transition-duration: 0s !important; ' +
+                 '-ms-transition-duration: 0s !important; ' +
+                 '-webkit-transition-duration: 0s !important; ' +
+
+                 'animation-delay: 0s !important; ' +
+                 '-o-animation-delay: 0s !important; ' +
+                 '-moz-animation-delay: 0s !important; ' +
+                 '-ms-animation-delay: 0s !important; ' +
+                 '-webkit-animation-delay: 0s !important; ' +
+
+                 'animation-duration: 0s !important; ' +
+                 '-o-animation-duration: 0s !important; ' +
+                 '-moz-animation-duration: 0s !important; ' +
+                 '-ms-animation-duration: 0s !important; ' +
+                 '-webkit-animation-duration: 0s !important; ' +
+
+                 '}';
+              document.head.appendChild(styleEl);
+              return true;
+            ''')
         except:
             pass # Will only work on HTML pages
 
@@ -714,7 +779,7 @@ class DriverBrowser(BasicBrowser):
         """Clicks on the element found by `locator`.
 
            :param locator: An instance of :class:`XPath` or a string containing an XPath expression.
-           :keyword wait: If False, don't wait_for_page_to_load after having typed into the input.
+           :keyword wait: If False, don't wait_for_page_to_load after having clicked the input.
         """
         self.wait_for_element_interactable(locator)
         self.find_element(locator).click()
@@ -758,7 +823,7 @@ class DriverBrowser(BasicBrowser):
         """
         xpath = six.text_type(locator)
         el = self.find_element(xpath)
-        return self.web_driver.execute_script('$(arguments[0]).focus()', el)
+        return self.web_driver.execute_script('arguments[0].focus();', el)
 
     @property
     def current_url(self):
@@ -791,12 +856,13 @@ class DriverBrowser(BasicBrowser):
         """
         return self.get_attribute(locator, 'value')
 
-    def execute_script(self, script):
+    def execute_script(self, script, *arguments):
         """Executes JavaScript in the browser.
 
-           :param script: A string containing the JavaScript to be executed.
+           :param script: A string containing the body of a JavaScript function to be executed.
+           :param arguments: Variable positional args passed into the function as an array named `arguments`.
         """
-        return self.web_driver.execute_script(script)
+        return self.web_driver.execute_script(script, *arguments)
 
     def switch_styling(self, javascript=True):
         """Switches styling for javascript enabled or javascript disabled.
@@ -845,7 +911,7 @@ class DriverBrowser(BasicBrowser):
         """
         return self.get_attribute(locator, 'href') is not None
     
-    def is_checked(self, locator): 
+    def is_checked(self, locator):
         """Answers whether the CheckBoxInput element found by `locator` is currently checked.
 
            :param locator: An instance of :class:`XPath` or a string containing an XPath expression.
@@ -858,7 +924,7 @@ class DriverBrowser(BasicBrowser):
            :param locator: An instance of :class:`XPath` or a string containing an XPath expression.
         """
         if not self.is_checked(locator):
-            self.click(locator)
+           self.click(self.checkbox_clickable_element(locator))
 
     def uncheck(self, locator):
         """Ensures the CheckBoxInput element found by `locator` is currently **not** checked.
@@ -866,7 +932,16 @@ class DriverBrowser(BasicBrowser):
            :param locator: An instance of :class:`XPath` or a string containing an XPath expression.
         """
         if self.is_checked(locator):
-            self.click(locator)
+            self.click(self.checkbox_clickable_element(locator))
+
+    def checkbox_clickable_element(self, locator):
+        if not self.is_visible(locator):
+            #it may be that the input type=checkbox may not be visible and represented by another visual element
+            #as my be in the case of for example with bootstrap .custom-checkbox.
+            #We simply click on the label, which checks the checkbox too.
+            label_for_input = XPath('(%s)/following-sibling::label' % locator.xpath)
+            return label_for_input
+        return locator
 
     def create_cookie(self, cookie_dict):
         """Creates a cookie from the given `cookie_dict`.
@@ -928,14 +1003,18 @@ class DriverBrowser(BasicBrowser):
             cropped = im.crop(bbox)
             cropped.save(output_file)
         except ImportError:
-            logging.warn('PILlow is not available, unable to crop screenshots')
+            logging.warning('PILlow is not available, unable to crop screenshots')
 
-    def press_tab(self, locator):
-        """Simulates the user pressing the tab key while the element at `locator` has focus.
+    def press_tab(self):
+        """Simulates the user pressing the tab key on element that is currently focussed.
 
-           :param locator: An instance of :class:`XPath` or a string containing an XPath expression.
+        .. versionchanged: 4.0
+           Changed to operate on the currently focussed element.
         """
-        self.find_element(locator).send_keys(Keys.TAB)
+        el = self.web_driver.switch_to.active_element
+        el.send_keys(Keys.TAB)
+        # To ensure the element gets blur event which for some reason does not always happen when pressing TAB:
+        self.web_driver.execute_script('if ( "undefined" !== typeof jQuery) {jQuery(arguments[0]).blur();};', el)
 
     def press_backspace(self, locator):
         """Simulates the user pressing the backspace key while the element at `locator` has focus.
