@@ -1,4 +1,4 @@
-# Copyright 2013-2016 Reahl Software Services (Pty) Ltd. All rights reserved.
+# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
 #
 #    This file is part of Reahl.
 #
@@ -32,20 +32,17 @@ from contextlib import contextmanager
 import logging
 import functools
 import pkg_resources
-import warnings
-import distutils
+
+from six.moves.http_client import CannotSendRequest
 
 from webob import Request
 from webob.exc import HTTPInternalServerError
 
-from six.moves.queue import Queue
-
 from watchdog.observers import Observer
 from watchdog.events import PatternMatchingEventHandler
-from watchdog.observers.polling import PollingObserver
 
 from reahl.component.exceptions import ProgrammerError
-from reahl.component.context import ExecutionContext
+from reahl.component.context import ExecutionContext, NoContextFound
 from reahl.component.config import StoredConfiguration
 from reahl.component.py3compat import ascii_as_bytes_or_str
 from reahl.component.shelltools import Executable
@@ -161,7 +158,7 @@ class SingleWSGIRequestHandler(simple_server.WSGIRequestHandler):
             return
         finally:
             self.request.settimeout(None)
-            
+
         self.raw_requestline = self.raw_requestline + self.rfile.readline(65536)
         if len(self.raw_requestline) > 65536:
             self.requestline = ''
@@ -179,9 +176,8 @@ class SingleWSGIRequestHandler(simple_server.WSGIRequestHandler):
         handler.request_handler = self      # backpointer for logging
         handler.run(self.server.get_app())
 
-        
-    def log_message(self, format, *args):
-        message = format % args
+    def log_message(self, message_format, *args):
+        message = message_format % args
         logging.getLogger(__name__).info(message)
 
 
@@ -284,7 +280,13 @@ class WebDriverHandler(object):
             def doit():
                 try:
                     started.set()
-                    r = self.original_execute(command, params)
+                    try:
+                        r = self.original_execute(command, params)
+                    except CannotSendRequest:
+                        # Retry in case the keep-alive connection state got mixed up
+                        # by, eg, the browser requesting a new url before all the
+                        # styleseets etc have loaded on the current one.
+                        r = self.original_execute(command, params)
                     results.append(r)
                 except Exception as e:
                     exceptions.append(e)
@@ -365,7 +367,7 @@ class SlaveProcess(object):
 
 class ServerSupervisor(PatternMatchingEventHandler):
     def __init__(self, slave_process_args, max_seconds_between_restarts, directories_to_monitor=['.']):
-        super(ServerSupervisor, self).__init__(ignore_patterns=['*.pyc','*.pyo', '*/__pycache__/*'], ignore_directories=True)
+        super(ServerSupervisor, self).__init__(ignore_patterns=['.git', '.floo', '*.pyc','*.pyo', '*/__pycache__/*'], ignore_directories=True)
         self.serving_process = SlaveProcess(sys.argv[0], slave_process_args)
         self.max_seconds_between_restarts = max_seconds_between_restarts
         self.directories_to_monitor = directories_to_monitor
@@ -462,28 +464,32 @@ class ReahlWebServer(object):
             self.httpsd = SSLCapableWSGIServer.make_server('', https_port, certfile, self.reahl_wsgi_app)
         except socket.error as ex:
             message = ('Caught socket.error: %s\nThis means that another process is using one of these ports: %s, %s. ' % (ex, port, https_port)) \
-                     +'\nIf this happens while running tests, it probably means that a browser client did not close its side of a connection to a previous server you had running - and that the server socket now sits in TIME_WAIT state. Is there perhaps a browser hanging around from a previous run? I have no idea how to fix this automatically... see http://hea-www.harvard.edu/~fine/Tech/addrinuse.html' \
-                      
+                     +'\nIf this happens while running tests, it probably means that a browser client did not close its side of a connection to a previous server you had running - and that the server socket now sits in TIME_WAIT state. Is there perhaps a browser hanging around from a previous run? I have no idea how to fix this automatically... see http://hea-www.harvard.edu/~fine/Tech/addrinuse.html'
+
             raise AssertionError(message)
 
-
-    def main_loop(self, context):
-        with context:
-            while self.running:
-                try:
-                    self.httpd.serve_async(in_separate_thread=self.in_separate_thread)
-                    self.httpsd.serve_async(in_separate_thread=self.in_separate_thread)
-                except:  
-                    # When running as a standalone server, we keep the server running, but else break so tests break
-                    if self.in_separate_thread and self.running:
-                        print(traceback.format_exc(), file=sys.stderr)
-                    else:
-                        raise
+    def main_loop(self, context=None):
+        if context:
+            context.install()
+        while self.running:
+            try:
+                self.httpd.serve_async(in_separate_thread=self.in_separate_thread)
+                self.httpsd.serve_async(in_separate_thread=self.in_separate_thread)
+            except:  
+                # When running as a standalone server, we keep the server running, but else break so tests break
+                if self.in_separate_thread and self.running:
+                    print(traceback.format_exc(), file=sys.stderr)
+                else:
+                    raise
 
     def start_thread(self):
         assert not self.running
         self.running = True
-        self.httpd_thread = Thread(target=functools.partial(self.main_loop, ExecutionContext.get_context()))
+        try:
+            context = ExecutionContext.get_context()
+        except NoContextFound:
+            context = None
+        self.httpd_thread = Thread(target=functools.partial(self.main_loop, context))
         self.httpd_thread.daemon = True
         self.httpd_thread.start()
 
@@ -501,7 +507,7 @@ class ReahlWebServer(object):
         finally:
             self.stop()
 
-    def start(self, in_separate_thread=True,  connect=False, in_seperate_thread=None):
+    def start(self, in_separate_thread=True,  connect=False):
         """Starts the webserver and web application.
         
            :keyword in_separate_thread: If False, the server handles requests in the same thread as your tests.
@@ -509,13 +515,7 @@ class ReahlWebServer(object):
            :keyword connect: If True, also connects to the database.
         """
         self.reahl_wsgi_app.start(connect=connect)
-
-        if in_seperate_thread:
-            warnings.warn('The in_seperate_thread keyword argument is deprecated, please use in_separate_thread instead.',
-               DeprecationWarning, stacklevel=2)
-            self.in_separate_thread = in_seperate_thread
-        else:
-            self.in_separate_thread = in_separate_thread
+        self.in_separate_thread = in_separate_thread
 
         if self.in_separate_thread:
             self.start_thread()
